/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "SrcInfo.h"
#include "Vector.h"
#include "Map.h"
#include <vector>

using namespace std;
using namespace FD;

class CvfdhCondensation;
   
DECLARE_NODE(CvfdhCondensation)
/*Node
 *
 * @name CvfdhCondensation
 * @category FDOPENCV_HARK:Segmentation
 * @description Output edge image
 *
 * @input_name IMAGEIN
 * @input_type any
 * @input_description Input image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @output_name SRCOUT
 * @output_type Vector<ObjectRef>
 * @output_description Source location. A type of an element of the vector is SrcInfo.
 *
 * @parameter_name N_PARTICLE
 * @parameter_type int
 * @parameter_value 4000
 * @parameter_description Number of paticles
 *
 * @parameter_name N_STATES
 * @parameter_type int
 * @parameter_value 4
 * @parameter_description Number of states
 *
 * @parameter_name SRC_TYPE
 * @parameter_type string
 * @parameter_value COG
 * @parameter_list COG:PARTICLE
 * @parameter_description Output source format
 *
 * @parameter_name ROI_SIZE
 * @parameter_type int
 * @parameter_value 5
 * @parameter_description Size of ROI for each source. If 0, default setting is loaded.
 *
 * @parameter_name TARGET_COLOR
 * @parameter_type object
 * @parameter_value <Vector<float> 255 255 255>
 * @parameter_description BGR color setting for tracking target. This is a 3D float vector which contains 0-255.
 *
 END*/
   
   
class CvfdhCondensation : public BufferedNode {

  int m_imageInID;
  int m_srcOutID;
  int n_particle;
  int n_stat;
  string src_type;
  int roi_size;
  Vector<float> target_color;

  int initialize_flag;

  float width;
  float height;
  CvConDensation *cond;
  CvMat *lowerBound;
  CvMat *upperBound;
      
public:
  CvfdhCondensation(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    m_imageInID = addInput("IMAGEIN");
    m_srcOutID = addOutput("SRCOUT");
    n_particle = dereference_cast<int>(parameters.get("N_PARTICLE"));
    n_stat = dereference_cast<int>(parameters.get("N_STATES"));
    src_type = object_cast<string>(parameters.get("SRC_TYPE"));
    roi_size = dereference_cast<int>(parameters.get("ROI_SIZE"));
    if((roi_size==0)&&(src_type == "PARTICLE")) roi_size = 10;
    target_color = object_cast<Vector<float> >(parameters.get("TARGET_COLOR"));
    initialize_flag = 0;
    inOrder = true;
  }      
      
  float calc_likelihood (IplImage * img, int x, int y, Vector<float>& color)
  {
    float b, g, r;
    float dist = 0.0, sigma = 50.0;
    
    b = img->imageData[img->widthStep * y + x * 3];       //B
    g = img->imageData[img->widthStep * y + x * 3 + 1];   //G
    r = img->imageData[img->widthStep * y + x * 3 + 2];   //R
    dist = sqrt ((color[0] - b) * (color[0] - b) + (color[1] - g) * (color[1] - g) + (color[2] - r) * (color[2] - r));
    
    return 1.0 / (sqrt (2.0 * CV_PI) * sigma) * expf (-dist * dist / (2.0 * sigma * sigma));
  }

  void particle_cog(CvConDensation *condtmp, vector<float>& cogtmp)
  {
    float lk;    
    cogtmp[0] = 0;
    cogtmp[1] = 0;
    lk = 0.0;
    for (int i = 0; i < condtmp->SamplesNum; i++) {
      cogtmp[0] += condtmp->flSamples[i][0] * cond->flConfidence[i];
      cogtmp[1] += condtmp->flSamples[i][1] * cond->flConfidence[i];
      lk += cond->flConfidence[i];
    }
    cogtmp[0] /= lk;
    cogtmp[1] /= lk;
  }

  void particle_max(CvConDensation *condtmp, vector<float>& maxtmp, vector<float>& mintmp)
  {
    maxtmp[0] = 0;
    maxtmp[1] = 0;
    mintmp[0] = 100000;
    mintmp[1] = 100000;
    for (int i = 0; i < condtmp->SamplesNum; i++) {
      if(cond->flConfidence[i] == 0.0) continue;
      if(maxtmp[0] < condtmp->flSamples[i][0]) maxtmp[0] = condtmp->flSamples[i][0];
      if(maxtmp[1] < condtmp->flSamples[i][1]) maxtmp[1] = condtmp->flSamples[i][1];
      if(mintmp[0] > condtmp->flSamples[i][0]) mintmp[0] = condtmp->flSamples[i][0];
      if(mintmp[1] > condtmp->flSamples[i][1]) mintmp[1] = condtmp->flSamples[i][1];
    }
  }

  void calculate(int output_id, int count, Buffer &out)
  {

    Vector<ObjectRef>& sources = *new Vector<ObjectRef>;
    (*(outputs[m_srcOutID].buffer))[count] = &sources;

    vector<IplImage*> image_vec;
    image_vec.resize(0);

    IplImage* img_in_tmp;
    ObjectRef in = getInput(m_imageInID, count);
    if (typeid(*in) == typeid(CvfdImage)) {
      RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);   
      if(imagePtr->getImage() == NULL){
	return;
      }
      img_in_tmp = cvCloneImage(imagePtr->getImage());
      image_vec.push_back(img_in_tmp);
    }else if (typeid(*in) == typeid(Map<int, ObjectRef>)) {
      RCPtr<Map<int, ObjectRef> > imageMapPtr = getInput(m_imageInID, count);
      Map<int, ObjectRef>::iterator it = imageMapPtr->begin();
      for (; it != imageMapPtr->end(); ++it) {
        CvfdImage& imagePtr = object_cast<CvfdImage>(it->second);
	img_in_tmp = cvCloneImage(imagePtr.getImage());
	image_vec.push_back(img_in_tmp);
      }
    }

    for(int cnt = 0; cnt < image_vec.size(); cnt++){
      IplImage* img_in;      
      img_in = cvCloneImage((IplImage*)image_vec[cnt]);
    
    // ---------------------------------------------------------------

    if(initialize_flag == 0){

      width  = (double)img_in->width;
      height = (double)img_in->height;
      
      cond = cvCreateConDensation (n_stat, 0, n_particle);
      
      lowerBound = cvCreateMat (4, 1, CV_32FC1);
      upperBound = cvCreateMat (4, 1, CV_32FC1);
    
      cvmSet (lowerBound, 0, 0, 0.0);
      cvmSet (lowerBound, 1, 0, 0.0);
      cvmSet (lowerBound, 2, 0, -10.0);
      cvmSet (lowerBound, 3, 0, -10.0);
      cvmSet (upperBound, 0, 0, width);
      cvmSet (upperBound, 1, 0, height);
      cvmSet (upperBound, 2, 0, 10.0);
      cvmSet (upperBound, 3, 0, 10.0);
      
      cvConDensInitSampleSet (cond, lowerBound, upperBound);
      
      cond->DynamMatr[0] = 1.0;
      cond->DynamMatr[1] = 0.0;
      cond->DynamMatr[2] = 1.0;
      cond->DynamMatr[3] = 0.0;
      cond->DynamMatr[4] = 0.0;
      cond->DynamMatr[5] = 1.0;
      cond->DynamMatr[6] = 0.0;
      cond->DynamMatr[7] = 1.0;
      cond->DynamMatr[8] = 0.0;
      cond->DynamMatr[9] = 0.0;
      cond->DynamMatr[10] = 1.0;
      cond->DynamMatr[11] = 0.0;
      cond->DynamMatr[12] = 0.0;
      cond->DynamMatr[13] = 0.0;
      cond->DynamMatr[14] = 0.0;
      cond->DynamMatr[15] = 1.0;
    
      cvRandInit (&(cond->RandS[0]), -25, 25, (int) cvGetTickCount ());
      cvRandInit (&(cond->RandS[1]), -25, 25, (int) cvGetTickCount ());
      cvRandInit (&(cond->RandS[2]), -5, 5, (int) cvGetTickCount ());
      cvRandInit (&(cond->RandS[3]), -5, 5, (int) cvGetTickCount ());
      
      initialize_flag = 1;
      
    }
    
    // ---------------------------------------------------------------
    // main routine : start

    IplImage* img_out;
    img_out = cvCloneImage (img_in);

    for (int i = 0; i < n_particle; i++) {
      int xx, yy;
      xx = (int) (cond->flSamples[i][0]);
      yy = (int) (cond->flSamples[i][1]);
      if (xx < 0 || xx >= width || yy < 0 || yy >= height) {
	cond->flConfidence[i] = 0.0;
      }
      else {
	cond->flConfidence[i] = calc_likelihood (img_in, xx, yy, target_color);
	if(src_type == "PARTICLE"){
	  RCPtr<SrcInfo> src(new SrcInfo);
	  src->roi_c[0] = xx;
	  src->roi_c[1] = yy;
	  src->roi_xy[0] = roi_size;
	  src->roi_xy[1] = roi_size;
	  src->img_size[0] = img_in->width;
	  src->img_size[1] = img_in->height;
	  sources.push_back(src);      
	}
      }
    }

    if(src_type == "COG"){    
      vector<float> cog;
      cog.resize(2);
      particle_cog(cond, cog);
      vector<float> maxpos, minpos;
      maxpos.resize(2);
      minpos.resize(2);      
      particle_max(cond, maxpos, minpos);
      RCPtr<SrcInfo> src(new SrcInfo);
      src->roi_c[0] = (int)cog[0];
      src->roi_c[1] = (int)cog[1];
      int roi_size_x = roi_size;
      int roi_size_y = roi_size;
      if(roi_size == 0){
	roi_size_x = (int)(maxpos[0] - minpos[0]);
	roi_size_y = (int)(maxpos[1] - minpos[1]);
      }
      src->roi_xy[0] = roi_size_x;
      src->roi_xy[1] = roi_size_y;
      src->img_size[0] = img_in->width;
      src->img_size[1] = img_in->height;
      sources.push_back(src);      
    }
    
    cvConDensUpdateByTime (cond);
            
    // main routine : finish
    // ---------------------------------------------------------------
         
    cvReleaseImage (&img_in);
    cvReleaseImage (&img_out);

    }

    cvReleaseImage(&img_in_tmp);

  }

      
  NO_ORDER_NODE_SPEEDUP(CvfdhCondensation)
};
