/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"

using namespace std;
using namespace FD;
   
class CvfdCmp;
   
DECLARE_NODE(CvfdCmp)
/*Node
 *
 * @name CvfdCmp
 * @category FDOPENCV_ORG:Arithmetic Logic Comparison
 * @description Performs per-element comparison of two images
 * dst(I) is set to 0xff (all '1'-bits) if the particular relation between the elements is true and 0 otherwise.
 * All the images must have the same type, except the destination, and the same size.
 *
 * @input_name IMAGEIN1
 * @input_description Input image1.
 * @input_type CvfdImage
 *
 * @input_name IMAGEIN2
 * @input_description Input image2.
 * @input_type CvfdImage
 *
 * @parameter_name CMP_OP
 * @parameter_type string
 * @parameter_value CV_CMP_EQ;CV_CMP_GT;CV_CMP_GE;CV_CMP_LT;CV_CMP_LE;CV_CMP_NE
 * @parameter_description : The flag specifying the relation between the elements to be checked
 *
 * @output_name IMAGEOUT
 * @output_description Output image
 * @output_type CvfdImage
 *
 *
 END*/
   
   
class CvfdCmp : public BufferedNode 
{
  //Input ID            
  int m_imageIn1ID;
  int m_imageIn2ID;
  //Output ID
  int m_imageOutID;
  //Parameters
  string m_cmpOp;
  //map of the line_type
  map<string,int> m_cmpOpMap;      
      
public:
  CvfdCmp(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    //add inputs
    m_imageIn1ID = addInput("IMAGEIN1");
    m_imageIn2ID = addInput("IMAGEIN2");
    //add outputs
    m_imageOutID = addOutput("IMAGEOUT");
    //Initialize parameters         
    m_cmpOp = object_cast<String>(parameters.get("CMP_OP"));         
         
    //Initialize the cmpOp map
    m_cmpOpMap["CV_CMP_EQ"] = CV_CMP_EQ;
    m_cmpOpMap["CV_CMP_GT"] = CV_CMP_GT;
    m_cmpOpMap["CV_CMP_GE"] = CV_CMP_GE;           
    m_cmpOpMap["CV_CMP_LT"] = CV_CMP_LT;
    m_cmpOpMap["CV_CMP_LE"] = CV_CMP_LE;
    m_cmpOpMap["CV_CMP_NE"] = CV_CMP_NE;          
  }      
      
  void calculate(int output_id, int count, Buffer &out)
  {
    //Read the inputs
    RCPtr<CvfdImage> imageIn1Ptr = getInput(m_imageIn1ID,count);   
    RCPtr<CvfdImage> imageIn2Ptr = getInput(m_imageIn2ID,count);
    //Handle
    CvfdImage* image = new CvfdImage(imageIn1Ptr->getImage());
    if(imageIn1Ptr->getImage()->nChannels == 1 && imageIn2Ptr->getImage()->nChannels == 1)
      {
	int status = cvGetErrMode();
	cvSetErrMode( CV_ErrModeSilent );
	__BEGIN__;
	OPENCV_CALL( cvCmp( imageIn1Ptr->getImage(), imageIn2Ptr->getImage(), image->getImage(), m_cmpOpMap[m_cmpOp] ));          
	__END__;
	cvSetErrMode( status );
	if( cvGetErrStatus() != CV_StsOk  )
	  {
	    throw new GeneralException("OPENCV - Error to compare the images: " +  CCHAR(cvErrorStr( cvGetErrStatus() )),__FILE__,__LINE__);
	  }
      }
    else
      {
	throw new GeneralException("OPENCV - Error to compare the images: they must be single-channel",__FILE__,__LINE__);
      }
         
    out[count] = ObjectRef(image);
  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdCmp)
};
