/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"

using namespace std;
using namespace FD;
   
class CvfdCreateImage;
   
DECLARE_NODE(CvfdCreateImage)
/*Node
 *
 * @name CvfdCreateImage
 * @category FDOPENCV_ORG:Image
 * @description Creates an image : header and allocates data
 *
 * @parameter_name WIDTH
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description Width of the image
 *
 * @parameter_name HEIGHT
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description Height of the image
 *
 * @parameter_name DEPTH
 * @parameter_type string
 * @parameter_value IPL_DEPTH_8U;IPL_DEPTH_8S;IPL_DEPTH_16U;IPL_DEPTH_16S;IPL_DEPTH_32S;IPL_DEPTH_32F;IPL_DEPTH_64F
 * @parameter_description Bit depth of image elements
 *
 * @parameter_name CHANNELS
 * @parameter_type int
 * @parameter_value 1
 * @parameter_description Number of channels per element(pixel). Can be 1, 2, 3 or 4
 *
 * @parameter_name VALUE
 * @parameter_type float
 * @parameter_value 0
 * @parameter_description Sets every element of image to given value
 *
 * @output_name IMAGE
 * @output_description The result image
 * @output_type CvfdImage
 *
 END*/
   
class CvfdCreateImage : public BufferedNode {
      
  //Output ID
  int m_imageID;
  //Parameter
  int m_width;
  int m_height;
  string m_depth;
  int m_channels;
  float m_value;
  //map of the depth
  map<string,int> m_depthMap;
      
public:
  CvfdCreateImage(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
         
    //add outputs
    m_imageID = addOutput("IMAGE");
         
    //Initialize parameters         
    m_width = dereference_cast<int>(parameters.get("WIDTH"));
    m_height = dereference_cast<int>(parameters.get("HEIGHT"));
    m_depth = object_cast<String>(parameters.get("DEPTH")); 
    m_channels = dereference_cast<int>(parameters.get("CHANNELS"));
    m_value = dereference_cast<float>(parameters.get("VALUE"));
         
    //Initialize lineType map
    m_depthMap["IPL_DEPTH_8U"] = IPL_DEPTH_8U;
    m_depthMap["IPL_DEPTH_8S"] = IPL_DEPTH_8S;
    m_depthMap["IPL_DEPTH_16U"] = IPL_DEPTH_16U;
    m_depthMap["IPL_DEPTH_16S"] = IPL_DEPTH_16S;
    m_depthMap["IPL_DEPTH_32S"] = IPL_DEPTH_32S;
    m_depthMap["IPL_DEPTH_32F"] = IPL_DEPTH_32F;
    m_depthMap["IPL_DEPTH_64F"] = IPL_DEPTH_64F;
         
  }
      
  void calculate(int output_id, int count, Buffer &out)
  {       
    //Handle of the inputs
    CvfdImage* image = new CvfdImage( cvSize(m_width,m_height) , m_depthMap[m_depth], m_channels );
    int status = cvGetErrMode();
    cvSetErrMode( CV_ErrModeSilent ); 
    __BEGIN__;
    OPENCV_CALL(cvSet(image->getImage(), cvScalarAll( (double)m_value )));
    __END__;
    cvSetErrMode( status );
         
    if( cvGetErrStatus() != CV_StsOk  )
      {        
	throw new GeneralException("OPENCV - Error to create the image: " +  CCHAR(cvErrorStr( cvGetErrStatus() )),__FILE__,__LINE__);
      }        
         
    out[count] = ObjectRef(image);
  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdCreateImage)
};
